from typing import NamedTuple

__all__ = ["version", "version_info", "version_js", "version_info_js"]


class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: str
    serial: int

    def __str__(self):
        v = f"{self.major}.{self.minor}.{self.micro}"
        level = self.releaselevel
        if level and level != "final":
            v = f"{v}{level[:1]}{self.serial}"
        return v


version = "3.0.0a0"

version_info = VersionInfo(3, 0, 0, "alpha", 0)

version_js = "0.5.0"

version_info_js = VersionInfo(0, 5, 0, "final", 0)



from pytest import mark  # type: ignore

from graphql import graphql
from graphql.type import (
    GraphQLField,
    GraphQLInputField,
    GraphQLInt,
    GraphQLObjectType,
    GraphQLSchema)

from graphql_relay import mutation_with_client_mutation_id


class Result:

    # noinspection PyPep8Naming
    def __init__(self, result, clientMutationId=None):
        self.clientMutationId = clientMutationId
        self.result = result


simple_mutation = mutation_with_client_mutation_id(
    'SimpleMutation',
    input_fields={},
    output_fields={
        'result': GraphQLField(GraphQLInt)
    },
    mutate_and_get_payload=lambda _info, **_input: Result(1)
)

simple_mutation_with_description = mutation_with_client_mutation_id(
    'SimpleMutationWithDescription',
    description='Simple Mutation Description',
    input_fields={},
    output_fields={
        'result': GraphQLField(GraphQLInt)
    },
    mutate_and_get_payload=lambda _info, **_input: Result(1)
)

simple_mutation_with_deprecation_reason = mutation_with_client_mutation_id(
    'SimpleMutationWithDeprecationReason',
    input_fields={},
    output_fields={
        'result': GraphQLField(GraphQLInt)
    },
    mutate_and_get_payload=lambda _info, **_input: Result(1),
    deprecation_reason='Just because'
)

# noinspection PyPep8Naming
simple_mutation_with_thunk_fields = mutation_with_client_mutation_id(
    'SimpleMutationWithThunkFields',
    input_fields=lambda: {
        'inputData': GraphQLInputField(GraphQLInt)
    },
    output_fields=lambda: {
        'result': GraphQLField(GraphQLInt)
    },
    mutate_and_get_payload=lambda _info, inputData, **_input: Result(inputData)
)


# noinspection PyPep8Naming
async def mutate_and_get_one_as_payload_async(_info, **_input):
    return Result(1)

simple_async_mutation = mutation_with_client_mutation_id(
    'SimpleAsyncMutation',
    input_fields={},
    output_fields={
        'result': GraphQLField(GraphQLInt)
    },
    mutate_and_get_payload=mutate_and_get_one_as_payload_async
)

simple_root_value_mutation = mutation_with_client_mutation_id(
    'SimpleRootValueMutation',
    input_fields={},
    output_fields={
        'result': GraphQLField(GraphQLInt)
    },
    mutate_and_get_payload=lambda info, **_input: info.root_value
)

query_type: GraphQLObjectType = GraphQLObjectType(
    'Query', lambda: {
        'query': GraphQLField(query_type)
    }
)

mutation_type = GraphQLObjectType(
    'Mutation',
    fields={
        'simpleMutation': simple_mutation,
        'simpleMutationWithDescription': simple_mutation_with_description,
        'simpleMutationWithDeprecationReason': simple_mutation_with_deprecation_reason,
        'simpleMutationWithThunkFields': simple_mutation_with_thunk_fields,
        'simpleAsyncMutation': simple_async_mutation,
        'simpleRootValueMutation': simple_root_value_mutation
    }
)

schema = GraphQLSchema(
    query=query_type,
    mutation=mutation_type
)


def describe_mutation_with_client_mutation_id():

    @mark.asyncio
    async def requires_an_argument():
        query = '''
          mutation M {
            simpleMutation {
              result
            }
          }
        '''
        assert await graphql(schema, query) == (
            None, [
                {
                    'message':
                        "Field 'simpleMutation' argument 'input'"
                        " of type 'SimpleMutationInput!' is required,"
                        " but it was not provided.",
                    'locations': [(3, 13)]
                }
            ]
        )

    @mark.asyncio
    async def returns_the_same_client_mutation_id():
        query = '''
          mutation M {
            simpleMutation(input: {clientMutationId: "abc"}) {
              result
              clientMutationId
            }
          }
        '''
        assert await graphql(schema, query) == (
            {
                'simpleMutation': {
                    'result': 1,
                    'clientMutationId': 'abc'
                }
            },
            None
        )

    @mark.asyncio
    async def supports_thunks_as_input_and_output_fields():
        query = '''
          mutation M {
            simpleMutationWithThunkFields(
                input: {inputData: 1234, clientMutationId: "abc"}) {
              result
              clientMutationId
            }
          }
        '''
        assert await graphql(schema, query) == (
            {
                'simpleMutationWithThunkFields': {
                    'result': 1234,
                    'clientMutationId': 'abc'
                }
            },
            None
        )

    @mark.asyncio
    async def supports_async_mutations():
        query = '''
          mutation M {
            simpleAsyncMutation(input: {clientMutationId: "abc"}) {
              result
              clientMutationId
            }
          }
        '''
        assert await graphql(schema, query) == (
            {
                'simpleAsyncMutation': {
                    'result': 1,
                    'clientMutationId': 'abc'
                }
            },
            None
        )

    @mark.asyncio
    async def can_access_root_value():
        query = '''
          mutation M {
            simpleRootValueMutation(input: {clientMutationId: "abc"}) {
              result
              clientMutationId
            }
          }
        '''
        assert await graphql(schema, query, root_value=Result(1)) == (
            {
                'simpleRootValueMutation': {
                    'result': 1,
                    'clientMutationId': 'abc'
                }
            },
            None
        )

    def describe_introspection():

        @mark.asyncio
        async def contains_correct_input():
            query = '''
              {
                __type(name: "SimpleMutationInput") {
                  name
                  kind
                  inputFields {
                    name
                    type {
                      name
                      kind
                    }
                  }
                }
              }
            '''
            assert await graphql(schema, query) == (
                {
                    '__type': {
                        'name': 'SimpleMutationInput',
                        'kind': 'INPUT_OBJECT',
                        'inputFields': [
                            {
                                'name': 'clientMutationId',
                                'type': {
                                    'name': None,
                                    'kind': 'NON_NULL',
                                }
                            }
                        ]
                    }
                },
                None
            )

        @mark.asyncio
        async def contains_correct_payload():
            query = '''
              {
                __type(name: "SimpleMutationPayload") {
                  name
                  kind
                  fields {
                    name
                    type {
                      name
                      kind
                    }
                  }
                }
              }
            '''
            assert await graphql(schema, query) == (
                {
                    '__type': {
                        'name': 'SimpleMutationPayload',
                        'kind': 'OBJECT',
                        'fields': [
                            {
                                'name': 'result',
                                'type': {
                                    'name': 'Int',
                                    'kind': 'SCALAR',
                                }
                            },
                            {
                                'name': 'clientMutationId',
                                'type': {
                                    'name': None,
                                    'kind': 'NON_NULL',
                                }
                            },
                        ]
                    }
                },
                None
            )

        @mark.asyncio
        async def contains_correct_field():
            query = '''
              {
                __schema {
                  mutationType {
                    fields {
                      name
                      args {
                        name
                        type {
                          name
                          kind
                          ofType {
                            name
                            kind
                          }
                        }
                      }
                      type {
                        name
                        kind
                      }
                    }
                  }
                }
              }
            '''
            assert await graphql(schema, query) == (
                {
                    '__schema': {
                        'mutationType': {
                            'fields': [
                                {
                                    'name': 'simpleMutation',
                                    'args': [
                                        {
                                            'name': 'input',
                                            'type': {
                                                'name': None,
                                                'kind': 'NON_NULL',
                                                'ofType': {
                                                    'name': 'SimpleMutationInput',
                                                    'kind': 'INPUT_OBJECT'
                                                }
                                            },
                                        }
                                    ],
                                    'type': {
                                        'name': 'SimpleMutationPayload',
                                        'kind': 'OBJECT',
                                    }
                                },
                                {
                                    'name': 'simpleMutationWithDescription',
                                    'args': [
                                        {
                                            'name': 'input',
                                            'type': {
                                                'name': None,
                                                'kind': 'NON_NULL',
                                                'ofType': {
                                                    'name': 'SimpleMutation'
                                                        'WithDescriptionInput',
                                                    'kind': 'INPUT_OBJECT',
                                                },
                                            },
                                        },
                                    ],
                                    'type': {
                                        'name': 'SimpleMutationWithDescriptionPayload',
                                        'kind': 'OBJECT',
                                    },
                                },
                                {
                                    'name': 'simpleMutationWithThunkFields',
                                    'args': [
                                        {
                                            'name': 'input',
                                            'type': {
                                                'name': None,
                                                'kind': 'NON_NULL',
                                                'ofType': {
                                                    'name': 'SimpleMutation'
                                                        'WithThunkFieldsInput',
                                                    'kind': 'INPUT_OBJECT'
                                                }
                                            },
                                        }
                                    ],
                                    'type': {
                                        'name': 'SimpleMutationWithThunkFieldsPayload',
                                        'kind': 'OBJECT',
                                    }
                                },
                                {
                                    'name': 'simpleAsyncMutation',
                                    'args': [
                                        {
                                            'name': 'input',
                                            'type': {
                                                'name': None,
                                                'kind': 'NON_NULL',
                                                'ofType': {
                                                    'name': 'SimpleAsyncMutationInput',
                                                    'kind': 'INPUT_OBJECT'
                                                }
                                            },
                                        }
                                    ],
                                    'type': {
                                        'name': 'SimpleAsyncMutationPayload',
                                        'kind': 'OBJECT',
                                    }
                                },
                                {
                                    'name': 'simpleRootValueMutation',
                                    'args': [
                                        {
                                            'name': 'input',
                                            'type': {
                                                'name': None,
                                                'kind': 'NON_NULL',
                                                'ofType': {
                                                    'name':
                                                        'SimpleRootValueMutationInput',
                                                    'kind': 'INPUT_OBJECT'
                                                }
                                            },
                                        }
                                    ],
                                    'type': {
                                        'name': 'SimpleRootValueMutationPayload',
                                        'kind': 'OBJECT',
                                    }
                                },
                            ]
                        }
                    }
                },
                None
            )

        @mark.asyncio
        async def contains_correct_descriptions():
            query = '''
              {
                __schema {
                  mutationType {
                    fields {
                      name
                      description
                    }
                  }
                }
              }
            '''
            assert await graphql(schema, query) == (
                {
                    '__schema': {
                        'mutationType': {
                            'fields': [
                                {
                                    'name': 'simpleMutation',
                                    'description': None,
                                },
                                {
                                    'name': 'simpleMutationWithDescription',
                                    'description': 'Simple Mutation Description',
                                },
                                {
                                    'name': 'simpleMutationWithThunkFields',
                                    'description': None,
                                },
                                {
                                    'name': 'simpleAsyncMutation',
                                    'description': None,
                                },
                                {
                                    'name': 'simpleRootValueMutation',
                                    'description': None,
                                },
                            ]
                        }
                    }
                },
                None
            )

        @mark.asyncio
        async def contains_correct_deprecation_reason():
            query = '''
              {
                __schema {
                  mutationType {
                    fields(includeDeprecated: true) {
                      name
                      isDeprecated
                      deprecationReason
                    }
                  }
                }
              }
            '''
            assert await graphql(schema, query) == (
                {
                    '__schema': {
                        'mutationType': {
                            'fields': [
                                {
                                    'name': 'simpleMutation',
                                    'isDeprecated': False,
                                    'deprecationReason': None,
                                },
                                {
                                    'name': 'simpleMutationWithDescription',
                                    'isDeprecated': False,
                                    'deprecationReason': None,
                                },
                                {
                                    'name': 'simpleMutationWithDeprecationReason',
                                    'isDeprecated': True,
                                    'deprecationReason': 'Just because',
                                },
                                {
                                    'name': 'simpleMutationWithThunkFields',
                                    'isDeprecated': False,
                                    'deprecationReason': None,
                                },
                                {
                                    'name': 'simpleAsyncMutation',
                                    'isDeprecated': False,
                                    'deprecationReason': None,
                                },
                                {
                                    'name': 'simpleRootValueMutation',
                                    'isDeprecated': False,
                                    'deprecationReason': None,
                                },
                            ]
                        }
                    }
                },
                None
            )



"""The graphql_relay package"""

# The graphql-relay and graphql-relay-js version info
from .version import version, version_info, version_js, version_info_js

# Types for creating connection types in the schema
from .connection.connectiontypes import (
    Connection, ConnectionArguments, ConnectionCursor, Edge, PageInfo)

# Helpers for creating connection types in the schema
from .connection.connection import (
    backward_connection_args, connection_args, connection_definitions,
    forward_connection_args, GraphQLConnectionDefinitions)

# Helpers for creating connections from arrays
from .connection.arrayconnection import (
    connection_from_array, connection_from_array_slice,
    cursor_for_object_in_connection,
    cursor_to_offset, get_offset_with_default, offset_to_cursor)

# Helper for creating mutations with client mutation IDs
from .mutation.mutation import mutation_with_client_mutation_id

# Helper for creating node definitions
from .node.node import node_definitions

#  Helper for creating plural identifying root fields
from .node.plural import plural_identifying_root_field

# Utilities for creating global IDs in systems that don't have them
from .node.node import (
    from_global_id, global_id_field, to_global_id)

# Deprecated functions from older graphql-relay-py versions
# noinspection PyProtectedMember,PyUnresolvedReferences,PyDeprecation
from .connection.arrayconnection import (  # noqa: F401
    connection_from_list, connection_from_list_slice)

__version__ = version
__version_info__ = version_info
__version_js__ = version_js
__version_info_js__ = version_info_js

__all__ = [
    'version', 'version_info', 'version_js', 'version_info_js',
    'Connection', 'ConnectionArguments', 'ConnectionCursor', 'Edge', 'PageInfo',
    'backward_connection_args', 'connection_args', 'connection_definitions',
    'forward_connection_args', 'GraphQLConnectionDefinitions',
    'connection_from_array', 'connection_from_array_slice',
    'cursor_for_object_in_connection', 'cursor_to_offset',
    'get_offset_with_default', 'offset_to_cursor',
    'mutation_with_client_mutation_id',
    'node_definitions',
    'plural_identifying_root_field',
    'from_global_id', 'global_id_field', 'to_global_id',
]



from typing import Any, Callable, NamedTuple

from graphql_relay.utils.base64 import base64, unbase64

from graphql.type import (
    GraphQLArgument,
    GraphQLNonNull,
    GraphQLID,
    GraphQLField,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLResolveInfo,
    GraphQLTypeResolver
)


class GraphQLNodeDefinitions(NamedTuple):

    node_interface: GraphQLInterfaceType
    node_field: GraphQLField
    nodes_field: GraphQLField


def node_definitions(
        id_fetcher: Callable[[str, GraphQLResolveInfo], Any],
        type_resolver: GraphQLTypeResolver = None) -> GraphQLNodeDefinitions:
    """
    Given a function to map from an ID to an underlying object, and a function
    to map from an underlying object to the concrete GraphQLObjectType it
    corresponds to, constructs a `Node` interface that objects can implement,
    and a field object to be used as a `node` root field.

    If the type_resolver is omitted, object resolution on the interface will be
    handled with the `is_type_of` method on object types, as with any GraphQL
    interface without a provided `resolve_type` method.
    """
    node_interface = GraphQLInterfaceType(
        'Node',
        description='An object with an ID',
        fields=lambda: {
            'id': GraphQLField(
                GraphQLNonNull(GraphQLID),
                description='The id of the object.')},
        resolve_type=type_resolver,
    )

    # noinspection PyShadowingBuiltins
    node_field = GraphQLField(
        node_interface,
        description='Fetches an object given its ID',
        args={
            'id': GraphQLArgument(
                GraphQLNonNull(GraphQLID),
                description='The ID of an object')},
        resolve=lambda _obj, info, id: id_fetcher(id, info)
    )

    nodes_field = GraphQLField(
        GraphQLNonNull(GraphQLList(node_interface)),
        args={
            'ids': GraphQLArgument(
                GraphQLNonNull(GraphQLList(GraphQLNonNull(GraphQLID))),
                description='The IDs of objects')},
        resolve=lambda _obj, info, ids: [id_fetcher(id_, info) for id_ in ids]
    )

    return GraphQLNodeDefinitions(node_interface, node_field, nodes_field)


class ResolvedGlobalId(NamedTuple):

    type: str
    id: str


def to_global_id(type_: str, id_: str) -> str:
    """
    Takes a type name and an ID specific to that type name, and returns a
    "global ID" that is unique among all types.
    """
    return base64(f'{type_}:{id_}')


def from_global_id(global_id: str) -> ResolvedGlobalId:
    """
    Takes the "global ID" created by to_global_id, and returns the type name and ID
    used to create it.
    """
    return ResolvedGlobalId(*unbase64(global_id).split(':', 1))


def global_id_field(
        type_name: str = None,
        id_fetcher: Callable[[Any, GraphQLResolveInfo], str] = None) -> GraphQLField:
    """
    Creates the configuration for an id field on a node, using `to_global_id` to
    construct the ID from the provided typename. The type-specific ID is fetched
    by calling id_fetcher on the object, or if not provided, by accessing the `id`
    attribute of the object, or the `id` if the object is a dict.
    """

    def resolve(obj: Any, info: GraphQLResolveInfo, **_args: Any) -> str:
        type_ = type_name or info.parent_type.name
        id_ = id_fetcher(obj, info) if id_fetcher else (
            obj['id'] if isinstance(obj, dict) else obj.id)
        return to_global_id(type_, id_)

    return GraphQLField(
        GraphQLNonNull(GraphQLID),
        description='The ID of an object',
        resolve=resolve
    )



from typing import Any, Dict, NamedTuple, Optional, Union

from pytest import fixture  # type: ignore

from graphql import graphql_sync as graphql
from graphql.type import (
    GraphQLField,
    GraphQLList,
    GraphQLInt,
    GraphQLObjectType,
    GraphQLResolveInfo,
    GraphQLSchema,
    GraphQLString)

from graphql_relay import from_global_id, global_id_field, node_definitions


@fixture(scope='module', params=['object_access', 'dict_access'])
def schema(request):
    """Run each test with object access and dict access."""
    use_dicts = request.param == 'dict_access'

    if use_dicts:

        class User(dict):
            pass

        class Photo(dict):
            pass

        class Post(dict):
            pass

    else:

        class User(NamedTuple):
            id: int
            name: str

        class Photo(NamedTuple):
            photo_id: int
            width: int

        class Post(NamedTuple):
            id: int
            text: str

    user_data = {
        '1': User(id=1, name='John Doe'),
        '2': User(id=2, name='Jane Smith'),
    }

    photo_data = {
        '1': Photo(photo_id=1, width=300),
        '2': Photo(photo_id=2, width=400),
    }

    post_data = {
        '1': Post(id=1, text='lorem'),
        '2': Post(id=2, text='ipsum'),
    }

    def get_node(
            global_id: str, info: GraphQLResolveInfo
    ) -> Optional[Union[User, Photo, Post, Dict]]:
        assert info.schema is schema
        type_, id_ = from_global_id(global_id)
        if type_ == 'User':
            return user_data[id_]
        if type_ == 'Photo':
            return photo_data[id_]
        if type_ == 'Post':
            return post_data[id_]
        return None

    def get_node_type(
            obj: Union[User, Photo], info: GraphQLResolveInfo,
            _type: Any) -> Optional[GraphQLObjectType]:
        assert info.schema is schema
        if 'name' in obj if use_dicts else isinstance(obj, User):
            return user_type
        if 'photo_id' in obj if use_dicts else isinstance(obj, Photo):
            return photo_type
        if 'text' in obj if use_dicts else isinstance(obj, Post):
            return post_type
        return None

    node_interface, node_field = node_definitions(get_node, get_node_type)[:2]

    user_type = GraphQLObjectType(
        'User',
        fields=lambda: {
            'id': global_id_field('User'),
            'name': GraphQLField(GraphQLString),
        },
        interfaces=[node_interface]
    )

    photo_type = GraphQLObjectType(
        'Photo',
        fields=lambda: {
            'id': global_id_field(
                'Photo',
                lambda obj, _info: obj['photo_id'] if use_dicts else obj.photo_id),
            'width': GraphQLField(GraphQLInt),
        },
        interfaces=[node_interface]
    )

    post_type = GraphQLObjectType(
        'Post',
        fields=lambda: {
            'id': global_id_field(),
            'text': GraphQLField(GraphQLString),
        },
        interfaces=[node_interface]
    )

    query_type = GraphQLObjectType(
        'Query',
        fields=lambda: {
            'node': node_field,
            'allObjects': GraphQLField(
                GraphQLList(node_interface),
                resolve=lambda _root, _info: [
                    user_data['1'], user_data['2'],
                    photo_data['1'], photo_data['2'],
                    post_data['1'], post_data['2']
                ]
            )
        }
    )

    schema = GraphQLSchema(
        query=query_type,
        types=[user_type, photo_type, post_type]
    )

    yield schema


def describe_global_id_fields():

    def gives_different_ids(schema):
        query = '''
        {
          allObjects {
            id
          }
        }
        '''
        assert graphql(schema, query) == (
            {
                'allObjects': [
                    {
                        'id': 'VXNlcjox'
                    },
                    {
                        'id': 'VXNlcjoy'
                    },
                    {
                        'id': 'UGhvdG86MQ=='
                    },
                    {
                        'id': 'UGhvdG86Mg=='
                    },
                    {
                        'id': 'UG9zdDox',
                    },
                    {
                        'id': 'UG9zdDoy',
                    },
                ]
            },
            None
        )

    def refetches_the_ids(schema):
        query = '''
        {
          user: node(id: "VXNlcjox") {
            id
            ... on User {
              name
            }
          },
          photo: node(id: "UGhvdG86MQ==") {
            id
            ... on Photo {
              width
            }
          }
          post: node(id: "UG9zdDox") {
            id
            ... on Post {
              text
            }
          }
        }
        '''
        assert graphql(schema, query) == (
            {
                'user': {
                    'id': 'VXNlcjox',
                    'name': 'John Doe'
                },
                'photo': {
                    'id': 'UGhvdG86MQ==',
                    'width': 300
                },
                'post': {
                    'id': 'UG9zdDox',
                    'text': 'lorem',
                },
            },
            None
        )



from inspect import isawaitable
from typing import Any, Callable

from graphql.error import GraphQLError
from graphql.pyutils import AwaitableOrValue, inspect
from graphql.type import (
    GraphQLArgument,
    GraphQLField,
    GraphQLFieldMap,
    GraphQLInputField,
    GraphQLInputFieldMap,
    GraphQLInputObjectType,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLResolveInfo,
    GraphQLString,
    Thunk
)


# Note: Contrary to the Javascript implementation of MutationFn,
# the context is passed as part of the GraphQLResolveInfo and any arguments
# are passed individually as keyword arguments.
MutationFnWithoutArgs = Callable[[GraphQLResolveInfo], AwaitableOrValue[Any]]
# Unfortunately there is currently no syntax to indicate optional or keyword
# arguments in Python, so we also allow any other Callable as a workaround:
MutationFn = Callable[..., AwaitableOrValue[Any]]


def resolve_maybe_thunk(thing_or_thunk: Thunk) -> Any:
    return thing_or_thunk() if callable(thing_or_thunk) else thing_or_thunk


def mutation_with_client_mutation_id(
        name: str,
        input_fields: Thunk[GraphQLInputFieldMap],
        output_fields: Thunk[GraphQLFieldMap],
        mutate_and_get_payload: MutationFn,
        description: str = None,
        deprecation_reason: str = None) -> GraphQLField:
    """
    Returns a GraphQLFieldConfig for the specified mutation.

    The input_fields and output_fields should not include `clientMutationId`,
    as this will be provided automatically.

    An input object will be created containing the input fields, and an
    object will be created containing the output fields.

    mutate_and_get_payload will receive a GraphQLResolveInfo as first argument,
    and the input fields as keyword arguments, and it should return an object
    (or a dict) with an attribute (or a key) for each output field.
    It may return synchronously or asynchronously.
    """
    def augmented_input_fields() -> GraphQLInputFieldMap:
        return dict(
            resolve_maybe_thunk(input_fields),
            clientMutationId=GraphQLInputField(
                GraphQLNonNull(GraphQLString)))

    def augmented_output_fields() -> GraphQLFieldMap:
        return dict(
            resolve_maybe_thunk(output_fields),
            clientMutationId=GraphQLField(
                GraphQLNonNull(GraphQLString)))

    output_type = GraphQLObjectType(
        name + 'Payload',
        fields=augmented_output_fields)

    input_type = GraphQLInputObjectType(
        name + 'Input',
        fields=augmented_input_fields)

    # noinspection PyShadowingBuiltins
    async def resolve(_root, info, input):
        payload = mutate_and_get_payload(info, **input)
        if isawaitable(payload):
            payload = await payload
        try:
            payload.clientMutationId = input['clientMutationId']
        except KeyError:
            raise GraphQLError(
                'Cannot set clientMutationId'
                f' in the payload object {inspect(payload)}.')
        return payload

    return GraphQLField(
        output_type,
        description=description,
        deprecation_reason=deprecation_reason,
        args={'input': GraphQLArgument(GraphQLNonNull(input_type))},
        resolve=resolve)



from typing import Any, Dict, List, NamedTuple, Optional

__all__ = [
    'Connection', 'ConnectionArguments', 'ConnectionCursor', 'Edge', 'PageInfo'
]


"""A type alias for cursors in this implementation."""
ConnectionCursor = str


class PageInfo(NamedTuple):
    """A type designed to be exposed as `PageInfo` over GraphQL."""

    startCursor: Optional[ConnectionCursor]
    endCursor: Optional[ConnectionCursor]
    hasPreviousPage: Optional[bool]
    hasNextPage: Optional[bool]


class Edge(NamedTuple):
    """A type designed to be exposed as a `Edge` over GraphQL."""

    node: Any
    cursor: ConnectionCursor


class Connection(NamedTuple):
    """A type designed to be exposed as a `Connection` over GraphQL."""

    edges: List[Edge]
    pageInfo: PageInfo


"""A type describing the arguments a connection field receives in GraphQL.

The following kinds of arguments are expected (all optional):

    before: ConnectionCursor
    after: ConnectionCursor
    first: int
    last: int
"""
ConnectionArguments = Dict[str, Any]



from pytest import mark  # type: ignore

from graphql import graphql

from .star_wars_schema import StarWarsSchema


@mark.asyncio
async def test_correctly_mutates_dataset():
    query = '''
      mutation AddBWingQuery($input: IntroduceShipInput!) {
        introduceShip(input: $input) {
          ship {
            id
            name
          }
          faction {
            name
          }
          clientMutationId
        }
      }
    '''
    params = {
        'input': {
            'shipName': 'B-Wing',
            'factionId': '1',
            'clientMutationId': 'abcde',
        }
    }
    expected = {
        'introduceShip': {
            'ship': {
                'id': 'U2hpcDo5',
                'name': 'B-Wing'
            },
            'faction': {
                'name': 'Alliance to Restore the Republic'
            },
            'clientMutationId': 'abcde',
        }
    }
    result = await graphql(StarWarsSchema, query, variable_values=params)
    assert result == (expected, None)



from typing import Any, NamedTuple

from graphql.type import (
    GraphQLArgument,
    GraphQLArgumentMap,
    GraphQLBoolean,
    GraphQLField,
    GraphQLFieldMap,
    GraphQLFieldResolver,
    GraphQLInt,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLString,
    Thunk
)

__all__ = [
    'connection_definitions',
    'forward_connection_args', 'backward_connection_args', 'connection_args',
    'GraphQLConnectionDefinitions'
]


# Returns a GraphQLArgumentMap appropriate to include on a field
# whose return type is a connection type with forward pagination.
forward_connection_args: GraphQLArgumentMap = {
    'after': GraphQLArgument(GraphQLString),
    'first': GraphQLArgument(GraphQLInt),
}

# Returns a GraphQLArgumentMap appropriate to include on a field
# whose return type is a connection type with backward pagination.
backward_connection_args: GraphQLArgumentMap = {
    'before': GraphQLArgument(GraphQLString),
    'last': GraphQLArgument(GraphQLInt),
}

# Returns a GraphQLArgumentMap appropriate to include on a field
# whose return type is a connection type with bidirectional pagination.
connection_args = {
    **forward_connection_args, **backward_connection_args
}


class GraphQLConnectionDefinitions(NamedTuple):
    edge_type: GraphQLObjectType
    connection_type: GraphQLObjectType


def resolve_maybe_thunk(thing_or_thunk: Thunk) -> Any:
    return thing_or_thunk() if callable(thing_or_thunk) else thing_or_thunk


def connection_definitions(
        node_type: GraphQLObjectType,
        name: str = None,
        resolve_node: GraphQLFieldResolver = None,
        resolve_cursor: GraphQLFieldResolver = None,
        edge_fields: Thunk[GraphQLFieldMap] = None,
        connection_fields: Thunk[GraphQLFieldMap] = None
) -> GraphQLConnectionDefinitions:
    """Return GraphQLObjectTypes for a connection with the given name.

    The nodes of the returned object types will be of the specified type.
    """
    name = name or node_type.name
    edge_fields = edge_fields or {}
    connection_fields = connection_fields or {}

    edge_type = GraphQLObjectType(
        name + 'Edge',
        description='An edge in a connection.',
        fields=lambda: {
            'node': GraphQLField(
                node_type,
                resolve=resolve_node,
                description='The item at the end of the edge',
            ),
            'cursor': GraphQLField(
                GraphQLNonNull(GraphQLString),
                resolve=resolve_cursor,
                description='A cursor for use in pagination',
            ),
            **resolve_maybe_thunk(edge_fields)})

    connection_type = GraphQLObjectType(
        name + 'Connection',
        description='A connection to a list of items.',
        fields=lambda: {
            'pageInfo': GraphQLField(
                GraphQLNonNull(page_info_type),
                description='The Information to aid in pagination'
            ),
            'edges': GraphQLField(
                GraphQLList(edge_type),
                description='A list of edges.',
            ),
            **resolve_maybe_thunk(connection_fields)})

    return GraphQLConnectionDefinitions(edge_type, connection_type)


# The common page info type used by all connections.
page_info_type = GraphQLObjectType(
    'PageInfo',
    description='Information about pagination in a connection.',
    fields=lambda: {
        'hasNextPage': GraphQLField(
            GraphQLNonNull(GraphQLBoolean),
            description='When paginating forwards, are there more items?',
        ),
        'hasPreviousPage': GraphQLField(
            GraphQLNonNull(GraphQLBoolean),
            description='When paginating backwards, are there more items?',
        ),
        'startCursor': GraphQLField(
            GraphQLString,
            description='When paginating backwards, the cursor to continue.',
        ),
        'endCursor': GraphQLField(
            GraphQLString,
            description='When paginating forwards, the cursor to continue.',
        ),
    }
)



from typing import List, NamedTuple

from pytest import mark  # type: ignore

from graphql import graphql
from graphql.type import (
    GraphQLField,
    GraphQLInt,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLString,
)

from graphql_relay import (
    backward_connection_args,
    connection_args,
    connection_definitions,
    connection_from_array,
    forward_connection_args)


class User(NamedTuple):
    name: str
    friends: List[int]


all_users = [
    User(name='Dan', friends=[1, 2, 3, 4]),
    User(name='Nick', friends=[0, 2, 3, 4]),
    User(name='Lee', friends=[0, 1, 3, 4]),
    User(name='Joe', friends=[0, 1, 2, 4]),
    User(name='Tim', friends=[0, 1, 2, 3]),
]

friend_connection: GraphQLObjectType
user_connection: GraphQLObjectType

user_type = GraphQLObjectType(
    'User',
    fields=lambda: {
        'name': GraphQLField(GraphQLString),
        'friends': GraphQLField(
            friend_connection,
            args=connection_args,
            resolve=lambda user, _info, **args:
            connection_from_array(user.friends, args),
        ),
        'friendsForward': GraphQLField(
            user_connection,
            args=forward_connection_args,
            resolve=lambda user, _info, **args:
            connection_from_array(user.friends, args),
        ),
        'friendsBackward': GraphQLField(
            user_connection,
            args=backward_connection_args,
            resolve=lambda user, _info, **args:
            connection_from_array(user.friends, args),
        ),
    },
)

friend_connection = connection_definitions(
    user_type,
    name='Friend',
    resolve_node=lambda edge, _info: all_users[edge.node],
    edge_fields=lambda: {
        'friendshipTime': GraphQLField(
            GraphQLString,
            resolve=lambda user_, info_: 'Yesterday'
        ),
    },
    connection_fields=lambda: {
        'totalCount': GraphQLField(
            GraphQLInt,
            resolve=lambda _user, _info: len(all_users) - 1
        ),
    }
).connection_type


user_connection = connection_definitions(
    user_type,
    resolve_node=lambda edge, _info: all_users[edge.node]
).connection_type


query_type = GraphQLObjectType(
    'Query',
    fields=lambda: {
        'user': GraphQLField(
            user_type,
            resolve=lambda _root, _info: all_users[0]
        ),
    }
)

schema = GraphQLSchema(query=query_type)


def describe_connection_definition():

    @mark.asyncio
    async def includes_connection_and_edge_fields():
        query = '''
          query FriendsQuery {
            user {
              friends(first: 2) {
                totalCount
                edges {
                  friendshipTime
                  node {
                    name
                  }
                }
              }
            }
          }
        '''
        assert await graphql(schema, query) == (
            {
                'user': {
                    'friends': {
                        'totalCount': 4,
                        'edges': [
                            {
                                'friendshipTime': 'Yesterday',
                                'node': {
                                    'name': 'Nick'
                                }
                            },
                            {
                                'friendshipTime': 'Yesterday',
                                'node': {
                                    'name': 'Lee'
                                }
                            },
                        ]
                    }
                }
            },
            None
        )

    @mark.asyncio
    async def works_with_forward_connection_args():
        query = '''
          query FriendsQuery {
            user {
              friendsForward(first: 2) {
                edges {
                  node {
                    name
                  }
                }
              }
            }
          }
        '''
        assert await graphql(schema, query) == (
            {
                'user': {
                    'friendsForward': {
                        'edges': [
                            {
                                'node': {
                                    'name': 'Nick'
                                }
                            },
                            {
                                'node': {
                                    'name': 'Lee'
                                }
                            },
                        ]
                    }
                }
            },
            None
        )

    @mark.asyncio
    async def works_with_backward_connection_args():
        query = '''
          query FriendsQuery {
            user {
              friendsBackward(last: 2) {
                edges {
                  node {
                    name
                  }
                }
              }
            }
          }
        '''
        assert await graphql(schema, query) == (
            {
                'user': {
                    'friendsBackward': {
                        'edges': [
                            {
                                'node': {
                                    'name': 'Joe'
                                }
                            },
                            {
                                'node': {
                                    'name': 'Tim'
                                }
                            },
                        ]
                    }
                }
            },
            None
        )



from typing import NamedTuple

from pytest import mark  # type: ignore

from graphql import graphql
from graphql.type import (
    GraphQLField,
    GraphQLObjectType,
    GraphQLResolveInfo,
    GraphQLSchema,
    GraphQLString)

from graphql_relay import plural_identifying_root_field

user_type = GraphQLObjectType(
    'User',
    fields=lambda: {
        'username': GraphQLField(GraphQLString),
        'url': GraphQLField(GraphQLString),
    }
)


class User(NamedTuple):
    username: str
    url: str


def resolve_single_input(info: GraphQLResolveInfo, username: str) -> User:
    assert info.schema is schema
    lang = info.context.lang
    url = f'www.facebook.com/{username}?lang={lang}'
    return User(username=username, url=url)


query_type = GraphQLObjectType(
    'Query', lambda: {
        'usernames': plural_identifying_root_field(
            'usernames',
            description='Map from a username to the user',
            input_type=GraphQLString,
            output_type=user_type,
            resolve_single_input=resolve_single_input)})

schema = GraphQLSchema(query=query_type)


class Context:
    lang = 'en'


def describe_plural_identifying_root_field():

    @mark.asyncio
    async def allows_fetching():
        query = '''
        {
          usernames(usernames:["dschafer", "leebyron", "schrockn"]) {
            username
            url
          }
        }
        '''
        assert await graphql(schema, query, context_value=Context()) == (
            {
                'usernames': [
                    {
                        'username': 'dschafer',
                        'url': 'www.facebook.com/dschafer?lang=en'
                    },
                    {
                        'username': 'leebyron',
                        'url': 'www.facebook.com/leebyron?lang=en'
                    },
                    {
                        'username': 'schrockn',
                        'url': 'www.facebook.com/schrockn?lang=en'
                    },
                ]
            },
            None
        )

    @mark.asyncio
    async def correctly_introspects():
        query = '''
        {
          __schema {
            queryType {
              fields {
                name
                args {
                  name
                  type {
                    kind
                    ofType {
                      kind
                      ofType {
                        kind
                        ofType {
                          name
                          kind
                        }
                      }
                    }
                  }
                }
                type {
                  kind
                  ofType {
                    name
                    kind
                  }
                }
              }
            }
          }
        }
        '''
        assert await graphql(schema, query) == (
            {
                '__schema': {
                    'queryType': {
                        'fields': [
                            {
                                'name': 'usernames',
                                'args': [
                                    {
                                        'name': 'usernames',
                                        'type': {
                                            'kind': 'NON_NULL',
                                            'ofType': {
                                                'kind': 'LIST',
                                                'ofType': {
                                                    'kind': 'NON_NULL',
                                                    'ofType': {
                                                        'name': 'String',
                                                        'kind': 'SCALAR',
                                                    }
                                                }
                                            }
                                        }
                                    }
                                ],
                                'type': {
                                    'kind': 'LIST',
                                    'ofType': {
                                        'name': 'User',
                                        'kind': 'OBJECT',
                                    }
                                }
                            }
                        ]
                    }
                }
            },
            None,
        )



from typing import Any, Callable

from graphql.type import (
    GraphQLArgument,
    GraphQLField,
    GraphQLInputType,
    GraphQLOutputType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLResolveInfo)


def plural_identifying_root_field(
        arg_name: str,
        input_type: GraphQLInputType,
        output_type: GraphQLOutputType,
        resolve_single_input: Callable[[GraphQLResolveInfo, str], Any],
        description: str = None) -> GraphQLField:
    if isinstance(input_type, GraphQLNonNull):
        input_type = input_type.of_type
    input_args = {arg_name: GraphQLArgument(
        GraphQLNonNull(GraphQLList(GraphQLNonNull(input_type))))}

    def resolve(_obj, info, **args):
        inputs = args[arg_name]
        return [resolve_single_input(info, input_) for input_ in inputs]

    return GraphQLField(
        GraphQLList(output_type),
        description=description,
        args=input_args,
        resolve=resolve)



from typing import NamedTuple

from pytest import mark  # type: ignore

from graphql import graphql
from graphql.type import (
    GraphQLField,
    GraphQLID,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLString)

from graphql_relay import node_definitions


class User(NamedTuple):
    id: str
    name: str


user_data = {
    '1': User(id='1', name='John Doe'),
    '2': User(id='2', name='Jane Smith'),
}

user_type: GraphQLObjectType

node_interface, node_field = node_definitions(
    lambda id_, _info: user_data[id_], lambda _obj, _info, _type: user_type)[:2]


user_type = GraphQLObjectType(
    'User', lambda: {
        'id': GraphQLField(GraphQLNonNull(GraphQLID)),
        'name': GraphQLField(GraphQLString),
    },
    interfaces=[node_interface]
)

query_type = GraphQLObjectType(
    'Query', lambda: {
        'node': node_field
    }
)

schema = GraphQLSchema(
    query=query_type,
    types=[user_type]
)


def describe_node_interface_and_fields_with_async_object_fetcher():

    @mark.asyncio
    async def gets_the_correct_id_for_users():
        query = '''
          {
            node(id: "1") {
              id
            }
          }
        '''
        assert await graphql(schema, query) == (
            {
                'node': {'id': '1'}
            },
            None
        )

    @mark.asyncio
    async def gets_the_correct_name_for_users():
        query = '''
          {
            node(id: "1") {
              id
              ... on User {
                name
              }
            }
          }
        '''
        assert await graphql(schema, query) == (
            {
                'node': {'id': '1', 'name': 'John Doe'}
            },
            None
        )



from typing import Any, NamedTuple, Optional, Union

from graphql import graphql_sync as graphql
from graphql.type import (
    GraphQLField,
    GraphQLID,
    GraphQLInt,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLResolveInfo,
    GraphQLSchema,
    GraphQLString)

from graphql_relay import node_definitions, to_global_id, from_global_id


class User(NamedTuple):
    id: str
    name: str


class Photo(NamedTuple):
    id: str
    width: int


user_data = {
    '1': User(id='1', name='John Doe'),
    '2': User(id='2', name='Jane Smith'),
}

photo_data = {
    '3': Photo(id='3', width=300),
    '4': Photo(id='4', width=400),
}


def get_node(id_: str, info: GraphQLResolveInfo) -> Optional[Union[User, Photo]]:
    assert info.schema is schema
    if id_ in user_data:
        return user_data[id_]
    if id_ in photo_data:
        return photo_data[id_]
    return None


def get_node_type(
        obj: Union[User, Photo], info: GraphQLResolveInfo,
        _type: Any) -> Optional[GraphQLObjectType]:
    assert info.schema is schema
    if obj.id in user_data:
        return user_type
    if obj.id in photo_data:
        return photo_type
    return None


node_interface, node_field, nodes_field = node_definitions(get_node, get_node_type)


user_type = GraphQLObjectType(
    'User', lambda: {
        'id': GraphQLField(GraphQLNonNull(GraphQLID)),
        'name': GraphQLField(GraphQLString),
    },
    interfaces=[node_interface]
)

photo_type = GraphQLObjectType(
    'Photo', lambda: {
        'id': GraphQLField(GraphQLNonNull(GraphQLID)),
        'width': GraphQLField(GraphQLInt),
    },
    interfaces=[node_interface]
)

query_type = GraphQLObjectType(
    'Query', lambda: {
        'node': node_field,
        'nodes': nodes_field
    }
)

schema = GraphQLSchema(
    query=query_type,
    types=[user_type, photo_type]
)


def describe_node_interface_and_fields():

    def describe_refetchability():

        def gets_the_correct_id_for_users():
            query = '''
              {
                node(id: "1") {
                  id
                }
              }
            '''
            assert graphql(schema, query) == (
                {
                    'node': {'id': '1'}
                },
                None
            )

        def gets_the_correct_name_for_users():
            query = '''
              {
                node(id: "1") {
                  id
                  ... on User {
                    name
                  }
                }
              }
            '''
            assert graphql(schema, query) == (
                {
                    'node': {'id': '1', 'name': 'John Doe'}
                },
                None
            )

        def gets_the_correct_width_for_photos():
            query = '''
              {
                node(id: "4") {
                  id
                  ... on Photo {
                    width
                  }
                }
              }
            '''
            assert graphql(schema, query) == (
                {
                    'node': {'id': '4', 'width': 400}
                },
                None
            )

        def gets_the_correct_typename_for_users():
            query = '''
              {
                node(id: "1") {
                  id
                  __typename
                }
              }
            '''
            assert graphql(schema, query) == (
                {
                    'node': {'id': '1', '__typename': 'User'}
                },
                None
            )

        def gets_the_correct_typename_for_photos():
            query = '''
              {
                node(id: "4") {
                  id
                  __typename
                }
              }
            '''
            assert graphql(schema, query) == (
                {
                    'node': {'id': '4', '__typename': 'Photo'}
                },
                None
            )

        def ignores_photo_fragments_on_user():
            query = '''
              {
                node(id: "1") {
                  id
                  ... on Photo {
                    width
                  }
                }
              }
            '''
            assert graphql(schema, query) == (
                {
                    'node': {'id': '1'}
                },
                None
            )

        def returns_null_for_bad_ids():
            query = '''
              {
                node(id: "5") {
                  id
                }
              }
            '''
            assert graphql(schema, query) == ({'node': None}, None)

        def returns_nulls_for_bad_ids():
            query = '''
              {
                nodes(ids: ["3", "5"]) {
                  id
                }
              }
            '''
            assert graphql(schema, query) == (
                {
                    'nodes': [{'id': '3'}, None],
                },
                None
            )

    def describe_introspection():

        def has_correct_node_interface():
            query = '''
              {
                __type(name: "Node") {
                  name
                  kind
                  fields {
                    name
                    type {
                      kind
                      ofType {
                        name
                        kind
                      }
                    }
                  }
                }
              }
            '''
            assert graphql(schema, query) == (
                {
                    '__type': {
                        'name': 'Node',
                        'kind': 'INTERFACE',
                        'fields': [
                            {
                                'name': 'id',
                                'type': {
                                    'kind': 'NON_NULL',
                                    'ofType': {
                                        'name': 'ID',
                                        'kind': 'SCALAR'
                                    }
                                }
                            }
                        ]
                    }
                },
                None
            )

        def has_correct_node_and_nodes_root_field():
            query = '''
              {
                __schema {
                  queryType {
                    fields {
                      name
                      type {
                        name
                        kind
                      }
                      args {
                        name
                        type {
                          kind
                          ofType {
                            name
                            kind
                          }
                        }
                      }
                    }
                  }
                }
              }
            '''
            assert graphql(schema, query) == (
                {
                    '__schema': {
                        'queryType': {
                            'fields': [
                                {
                                    'name': 'node',
                                    'type': {
                                        'name': 'Node',
                                        'kind': 'INTERFACE'
                                    },
                                    'args': [
                                        {
                                            'name': 'id',
                                            'type': {
                                                'kind': 'NON_NULL',
                                                'ofType': {
                                                    'name': 'ID',
                                                    'kind': 'SCALAR'
                                                }
                                            }
                                        }
                                    ]
                                },
                                {
                                    'name': 'nodes',
                                    'type': {
                                        'name': None,
                                        'kind': 'NON_NULL',
                                    },
                                    'args': [
                                        {
                                            'name': 'ids',
                                            'type': {
                                                'kind': 'NON_NULL',
                                                'ofType': {
                                                    'name': None,
                                                    'kind': 'LIST',
                                                },
                                            },
                                        },
                                    ],
                                },
                            ]
                        }
                    }
                },
                None
            )


def describe_convert_global_ids():

    def to_global_id_converts_unicode_strings_correctly():
        my_unicode_id = 'ûñö'
        g_id = to_global_id('MyType', my_unicode_id)
        assert g_id == 'TXlUeXBlOsO7w7HDtg=='

        my_unicode_id = '\u06ED'
        g_id = to_global_id('MyType', my_unicode_id)
        assert g_id == 'TXlUeXBlOtut'

    def from_global_id_converts_unicode_strings_correctly():
        my_unicode_id = 'ûñö'
        my_type, my_id = from_global_id('TXlUeXBlOsO7w7HDtg==')
        assert my_type == 'MyType'
        assert my_id == my_unicode_id

        my_unicode_id = '\u06ED'
        my_type, my_id = from_global_id('TXlUeXBlOtut')
        assert my_type == 'MyType'
        assert my_id == my_unicode_id



from base64 import b64encode, b64decode

__all__ = ['base64', 'unbase64']


def base64(s: str) -> str:
    """"Encode the string s using Base64."""
    return b64encode(s.encode('utf-8')).decode('utf-8')


def unbase64(s: str) -> str:
    """"Decode the string s using Base64."""
    return b64decode(s).decode('utf-8')



from graphql_relay.utils import base64, unbase64


example_unicode = 'Some examples: ❤😀'
example_base64 = 'U29tZSBleGFtcGxlczog4p2k8J+YgA=='


def describe_base64_conversion():

    def converts_from_unicode_to_base64():
        assert base64(example_unicode) == example_base64

    def converts_from_base_64_to_unicode():
        assert unbase64(example_base64) == example_unicode



from re import search
from setuptools import setup, find_packages

with open('src/graphql_relay/version.py') as version_file:
    version = search('version = "(.*)"', version_file.read()).group(1)

with open('README.md') as readme_file:
    readme = readme_file.read()

setup(
    name='graphql-relay',
    version=version,
    description='Relay library for graphql-core-next',
    long_description=readme,
    long_description_content_type='text/markdown',
    keywords='graphql relay api',
    url='https://github.com/graphql-python/graphql-relay-py',
    author='Syrus Akbary',
    author_email='me@syrusakbary.com',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    install_requires=['graphql-core>=3.0.0a0'],
    python_requires='>=3.6,<4',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    # PEP-561: https://www.python.org/dev/peps/pep-0561/
    package_data={"graphql_relay": ["py.typed"]},
    include_package_data=True,
    zip_safe=False,
)



import re

import graphql_relay
from graphql_relay import version, version_info, version_js, version_info_js

_re_version = re.compile(r"(\d+)\.(\d+)\.(\d+)(?:(a|b|c)(\d+))?$")


def test_module_version():
    assert graphql_relay.__version__ == version
    assert graphql_relay.version == version


def test_module_version_info():
    assert graphql_relay.__version_info__ == version_info
    assert graphql_relay.version_info == version_info


def test_module_version_js():
    assert graphql_relay.__version_js__ == version_js
    assert graphql_relay.version_js == version_js


def test_module_version_info_js():
    assert graphql_relay.__version_info_js__ == version_info_js
    assert graphql_relay.version_info_js == version_info_js


def test_version():
    assert isinstance(version, str)
    assert _re_version.match(version)


def test_version_info():
    assert isinstance(version_info, tuple)
    assert str(version_info) == version
    groups = _re_version.match(version).groups()
    assert version_info.major == int(groups[0])
    assert version_info.minor == int(groups[1])
    assert version_info.micro == int(groups[2])
    if groups[3] is None:
        assert groups[4] is None
    else:
        assert version_info.releaselevel[:1] == groups[3]
        assert version_info.serial == int(groups[4])


def test_version_js():
    assert isinstance(version_js, str)
    assert _re_version.match(version_js)


def test_version_info_js():
    assert isinstance(version_info_js, tuple)
    assert str(version_info_js) == version_js
    groups = _re_version.match(version_js).groups()
    assert version_info_js.major == int(groups[0])
    assert version_info_js.minor == int(groups[1])
    assert version_info_js.micro == int(groups[2])
    if groups[3] is None:
        assert groups[4] is None
    else:
        assert version_info_js.releaselevel[:1] == groups[3]
        assert version_info_js.serial == int(groups[4])



from typing import cast, Sequence

from pytest import deprecated_call, raises  # type: ignore

from graphql_relay import (
    connection_from_array,
    connection_from_array_slice,
    cursor_for_object_in_connection,
    Connection, Edge, PageInfo)

# noinspection PyProtectedMember
from graphql_relay import connection_from_list, connection_from_list_slice


def describe_connection_from_array():
    letters = ['A', 'B', 'C', 'D', 'E']

    def describe_basic_slicing():

        def returns_all_elements_without_filters():
            c = connection_from_array(letters, {})
            assert c == Connection(
                edges=[
                    Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA='),
                    Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                    Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
                    Edge(node='E', cursor='YXJyYXljb25uZWN0aW9uOjQ='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjQ=',
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )

        def respects_a_smaller_first():
            c = connection_from_array(letters, dict(first=2))
            assert c == Connection(
                edges=[
                    Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA='),
                    Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjE=',
                    hasPreviousPage=False,
                    hasNextPage=True,
                )
            )

        def respects_an_overly_large_first():
            c = connection_from_array(letters, dict(first=10))
            assert c == Connection(
                edges=[
                    Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA='),
                    Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                    Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
                    Edge(node='E', cursor='YXJyYXljb25uZWN0aW9uOjQ='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjQ=',
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )

        def respects_a_smaller_last():
            c = connection_from_array(letters, dict(last=2))
            assert c == Connection(
                edges=[
                    Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
                    Edge(node='E', cursor='YXJyYXljb25uZWN0aW9uOjQ='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjM=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjQ=',
                    hasPreviousPage=True,
                    hasNextPage=False,
                )
            )

        def respects_an_overly_large_last():
            c = connection_from_array(letters, dict(last=10))
            assert c == Connection(
                edges=[
                    Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA='),
                    Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                    Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
                    Edge(node='E', cursor='YXJyYXljb25uZWN0aW9uOjQ='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjQ=',
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )

    def describe_pagination():

        def respects_first_and_after():
            c = connection_from_array(letters, dict(
                first=2, after='YXJyYXljb25uZWN0aW9uOjE='))
            assert c == Connection(
                edges=[
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                    Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjI=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjM=',
                    hasPreviousPage=False,
                    hasNextPage=True,
                )
            )

        def respects_first_and_after_with_long_first():
            c = connection_from_array(letters, dict(
                first=10, after='YXJyYXljb25uZWN0aW9uOjE='))
            assert c == Connection(
                edges=[
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                    Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
                    Edge(node='E', cursor='YXJyYXljb25uZWN0aW9uOjQ='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjI=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjQ=',
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )

        def respects_last_and_before():
            c = connection_from_array(letters, dict(
                last=2, before='YXJyYXljb25uZWN0aW9uOjM='))
            assert c == Connection(
                edges=[
                    Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjE=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjI=',
                    hasPreviousPage=True,
                    hasNextPage=False,
                )
            )

        def respects_last_and_before_with_long_last():
            c = connection_from_array(
                letters, dict(last=10, before='YXJyYXljb25uZWN0aW9uOjM='))
            assert c == Connection(
                edges=[
                    Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA='),
                    Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjI=',
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )

        def respects_first_and_after_and_before_too_few():
            c = connection_from_array(letters, dict(
                first=2,
                after='YXJyYXljb25uZWN0aW9uOjA=',
                before='YXJyYXljb25uZWN0aW9uOjQ=',
            ))
            assert c == Connection(
                edges=[
                    Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjE=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjI=',
                    hasPreviousPage=False,
                    hasNextPage=True,
                )
            )

        def respects_first_and_after_and_before_too_many():
            c = connection_from_array(letters, dict(
                first=4,
                after='YXJyYXljb25uZWN0aW9uOjA=',
                before='YXJyYXljb25uZWN0aW9uOjQ=',
            ))
            assert c == Connection(
                edges=[
                    Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                    Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjE=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjM=',
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )

        def respects_first_and_after_and_before_exactly_right():
            c = connection_from_array(letters, dict(
                first=3,
                after='YXJyYXljb25uZWN0aW9uOjA=',
                before='YXJyYXljb25uZWN0aW9uOjQ=',
            ))
            assert c == Connection(
                edges=[
                    Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                    Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjE=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjM=',
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )

        def respects_last_and_after_and_before_too_few():
            c = connection_from_array(letters, dict(
                last=2,
                after='YXJyYXljb25uZWN0aW9uOjA=',
                before='YXJyYXljb25uZWN0aW9uOjQ=',
            ))
            assert c == Connection(
                edges=[
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                    Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjI=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjM=',
                    hasPreviousPage=True,
                    hasNextPage=False,
                )
            )

        def respects_last_and_after_and_before_too_many():
            c = connection_from_array(letters, dict(
                last=4,
                after='YXJyYXljb25uZWN0aW9uOjA=',
                before='YXJyYXljb25uZWN0aW9uOjQ=',
            ))
            assert c == Connection(
                edges=[
                    Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                    Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjE=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjM=',
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )

        def respects_last_and_after_and_before_exactly_right():
            c = connection_from_array(letters, dict(
                last=3,
                after='YXJyYXljb25uZWN0aW9uOjA=',
                before='YXJyYXljb25uZWN0aW9uOjQ=',
            ))
            assert c == Connection(
                edges=[
                    Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                    Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjE=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjM=',
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )

    def describe_cursor_edge_cases():

        def throws_an_error_if_first_smaller_than_zero():
            with raises(ValueError) as exc_info:
                connection_from_array(letters, dict(first=-1))
            assert str(exc_info.value) == (
                "Argument 'first' must be a non-negative integer.")

        def throws_an_error_if_last_smaller_than_zero():
            with raises(ValueError) as exc_info:
                connection_from_array(letters, dict(last=-1))
            assert str(exc_info.value) == (
                "Argument 'last' must be a non-negative integer.")

        def returns_all_elements_if_cursors_are_invalid():
            c = connection_from_array(letters, dict(
                before='invalid', after='invalid'))
            assert c == Connection(
                edges=[
                    Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA='),
                    Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                    Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
                    Edge(node='E', cursor='YXJyYXljb25uZWN0aW9uOjQ='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjQ=',
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )

        def returns_all_elements_if_cursors_are_on_the_outside():
            c = connection_from_array(letters, dict(
                before='YXJyYXljb25uZWN0aW9uOjYK',
                after='YXJyYXljb25uZWN0aW9uOi0xCg=='
            ))
            assert c == Connection(
                edges=[
                    Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA='),
                    Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                    Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
                    Edge(node='E', cursor='YXJyYXljb25uZWN0aW9uOjQ='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjQ=',
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )

        def returns_no_elements_if_cursors_cross():
            c = connection_from_array(letters, dict(
                before='YXJyYXljb25uZWN0aW9uOjI=',
                after='YXJyYXljb25uZWN0aW9uOjQ='
            ))
            assert c == Connection(
                edges=[],
                pageInfo=PageInfo(
                    startCursor=None,
                    endCursor=None,
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )

    def describe_cursor_for_object_in_connection():

        def returns_an_edges_cursor_given_an_array_and_a_member_object():
            letter_b_cursor = cursor_for_object_in_connection(letters, 'B')
            assert letter_b_cursor == 'YXJyYXljb25uZWN0aW9uOjE='

        def returns_null_given_an_array_and_a_non_member_object():
            letter_f_cursor = cursor_for_object_in_connection(letters, 'F')
            assert letter_f_cursor is None

        def describe_extended_functionality():
            """Test functionality that is not part of graphql-relay-js."""

            def returns_an_edges_cursor_given_an_array_without_index_method():
                class LettersWithoutIndex:
                    __getitem__ = letters.__getitem__

                letters_without_index = LettersWithoutIndex()

                with raises(AttributeError):
                    cast(Sequence, letters_without_index).index('B')

                letter_b_cursor = cursor_for_object_in_connection(
                    cast(Sequence, letters_without_index), 'B')
                assert letter_b_cursor == 'YXJyYXljb25uZWN0aW9uOjE='

    def describe_extended_functionality():
        """Test functionality that is not part of graphql-relay-js."""

        def does_not_require_args():
            c = connection_from_array(letters)
            assert c == Connection(
                edges=[
                    Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA='),
                    Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                    Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
                    Edge(node='E', cursor='YXJyYXljb25uZWN0aW9uOjQ='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjQ=',
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )

        def uses_default_connection_types():
            connection = connection_from_array(letters[:1])
            assert isinstance(connection, Connection)
            edge = connection.edges[0]
            assert isinstance(edge, Edge)
            assert len(connection.edges) == 1
            assert edge == Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA=')
            page_info = connection.pageInfo
            assert isinstance(page_info, PageInfo)
            assert page_info == PageInfo(
                startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                endCursor='YXJyYXljb25uZWN0aW9uOjA=',
                hasPreviousPage=False,
                hasNextPage=False)

        def accepts_custom_connection_type():
            class CustomConnection:
                # noinspection PyPep8Naming
                def __init__(self, edges, pageInfo):
                    self.edges = edges
                    self.page_info = pageInfo

            connection = connection_from_array(
                letters[:1], connection_type=CustomConnection)
            assert isinstance(connection, CustomConnection)
            edge = connection.edges[0]
            assert isinstance(edge, Edge)
            assert len(connection.edges) == 1
            assert edge == Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA=')
            page_info = connection.page_info
            assert isinstance(page_info, PageInfo)
            assert page_info == PageInfo(
                startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                endCursor='YXJyYXljb25uZWN0aW9uOjA=',
                hasPreviousPage=False,
                hasNextPage=False)

        def accepts_custom_edge_type():
            class CustomEdge:
                def __init__(self, node, cursor):
                    self.node = node
                    self.cursor = cursor

            connection = connection_from_array(letters[:1], edge_type=CustomEdge)
            assert isinstance(connection, Connection)
            assert isinstance(connection.edges, list)
            assert len(connection.edges) == 1
            edge = connection.edges[0]
            assert isinstance(edge, CustomEdge)
            assert edge.node == 'A'
            assert edge.cursor == 'YXJyYXljb25uZWN0aW9uOjA='
            page_info = connection.pageInfo
            assert isinstance(page_info, PageInfo)
            assert page_info == PageInfo(
                startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                endCursor='YXJyYXljb25uZWN0aW9uOjA=',
                hasPreviousPage=False,
                hasNextPage=False)

        def accepts_custom_page_info_type():
            class CustomPageInfo:
                # noinspection PyPep8Naming
                def __init__(
                        self, startCursor, endCursor, hasPreviousPage, hasNextPage):
                    self.startCursor = startCursor
                    self.endCursor = endCursor
                    self.hasPreviousPage = hasPreviousPage
                    self.hasNextPage = hasNextPage

            connection = connection_from_array(
                letters[:1], page_info_type=CustomPageInfo)
            assert isinstance(connection, Connection)
            assert isinstance(connection.edges, list)
            assert len(connection.edges) == 1
            edge = connection.edges[0]
            assert isinstance(edge, Edge)
            assert edge == Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA=')
            page_info = connection.pageInfo
            assert isinstance(page_info, CustomPageInfo)
            assert page_info.startCursor == 'YXJyYXljb25uZWN0aW9uOjA='
            assert page_info.endCursor == 'YXJyYXljb25uZWN0aW9uOjA='
            assert page_info.hasPreviousPage is False
            assert page_info.hasNextPage is False

        def provides_deprecated_connection_from_list():
            with deprecated_call():
                # noinspection PyDeprecation
                c = connection_from_list(
                    letters[:1], args={}, connection_type=Connection,
                    edge_type=Edge, pageinfo_type=PageInfo)
            assert c == Connection(
                edges=[
                    Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )


def describe_connection_from_array_slice():
    letters = ['A', 'B', 'C', 'D', 'E']

    def works_with_a_just_right_array_slice():
        c = connection_from_array_slice(
            letters[1:3],
            dict(
                first=2,
                after='YXJyYXljb25uZWN0aW9uOjA=',
            ),
            slice_start=1, array_length=5
        )
        assert c == Connection(
            edges=[
                Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
            ],
            pageInfo=PageInfo(
                startCursor='YXJyYXljb25uZWN0aW9uOjE=',
                endCursor='YXJyYXljb25uZWN0aW9uOjI=',
                hasPreviousPage=False,
                hasNextPage=True,
            )
        )

    def works_with_an_oversized_array_slice_left_side():
        c = connection_from_array_slice(
            letters[0:3],
            dict(
                first=2,
                after='YXJyYXljb25uZWN0aW9uOjA=',
            ),
            slice_start=0, array_length=5
        )
        assert c == Connection(
            edges=[
                Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
            ],
            pageInfo=PageInfo(
                startCursor='YXJyYXljb25uZWN0aW9uOjE=',
                endCursor='YXJyYXljb25uZWN0aW9uOjI=',
                hasPreviousPage=False,
                hasNextPage=True,
            )
        )

    def works_with_an_oversized_array_slice_right_side():
        c = connection_from_array_slice(
            letters[2:4],
            dict(
                first=1,
                after='YXJyYXljb25uZWN0aW9uOjE=',
            ),
            slice_start=2, array_length=5
        )
        assert c == Connection(
            edges=[
                Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
            ],
            pageInfo=PageInfo(
                startCursor='YXJyYXljb25uZWN0aW9uOjI=',
                endCursor='YXJyYXljb25uZWN0aW9uOjI=',
                hasPreviousPage=False,
                hasNextPage=True,
            )
        )

    def works_with_an_oversized_array_slice_both_sides():
        c = connection_from_array_slice(
            letters[1:4],
            dict(
                first=1,
                after='YXJyYXljb25uZWN0aW9uOjE=',
            ),
            slice_start=1, array_length=5
        )
        assert c == Connection(
            edges=[
                Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
            ],
            pageInfo=PageInfo(
                startCursor='YXJyYXljb25uZWN0aW9uOjI=',
                endCursor='YXJyYXljb25uZWN0aW9uOjI=',
                hasPreviousPage=False,
                hasNextPage=True,
            )
        )

    def works_with_an_undersized_array_slice_left_side():
        c = connection_from_array_slice(
            letters[3:5],
            dict(
                first=3,
                after='YXJyYXljb25uZWN0aW9uOjE=',
            ),
            slice_start=3, array_length=5
        )
        assert c == Connection(
            edges=[
                Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
                Edge(node='E', cursor='YXJyYXljb25uZWN0aW9uOjQ='),
            ],
            pageInfo=PageInfo(
                startCursor='YXJyYXljb25uZWN0aW9uOjM=',
                endCursor='YXJyYXljb25uZWN0aW9uOjQ=',
                hasPreviousPage=False,
                hasNextPage=False,
            )
        )

    def works_with_an_undersized_array_slice_right_side():
        c = connection_from_array_slice(
            letters[2:4],
            dict(
                first=3,
                after='YXJyYXljb25uZWN0aW9uOjE=',
            ),
            slice_start=2, array_length=5
        )
        assert c == Connection(
            edges=[
                Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
            ],
            pageInfo=PageInfo(
                startCursor='YXJyYXljb25uZWN0aW9uOjI=',
                endCursor='YXJyYXljb25uZWN0aW9uOjM=',
                hasPreviousPage=False,
                hasNextPage=True,
            )
        )

    def works_with_an_undersized_array_slice_both_sides():
        c = connection_from_array_slice(
            letters[3:4],
            dict(
                first=3,
                after='YXJyYXljb25uZWN0aW9uOjE=',
            ),
            slice_start=3, array_length=5
        )
        assert c == Connection(
            edges=[
                Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
            ],
            pageInfo=PageInfo(
                startCursor='YXJyYXljb25uZWN0aW9uOjM=',
                endCursor='YXJyYXljb25uZWN0aW9uOjM=',
                hasPreviousPage=False,
                hasNextPage=True,
            )
        )

    def describe_extended_functionality():
        """Test functionality that is not part of graphql-relay-js."""

        def does_not_require_args():
            c = connection_from_array_slice(
                letters, slice_start=0, array_length=5
            )
            assert c == Connection(
                edges=[
                    Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA='),
                    Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                    Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
                    Edge(node='E', cursor='YXJyYXljb25uZWN0aW9uOjQ='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjQ=',
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )

        def uses_zero_as_default_for_slice_start():
            c = connection_from_array_slice(
                letters[:1], dict(first=1), array_length=5
            )
            assert c == Connection(
                edges=[
                    Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    hasPreviousPage=False,
                    hasNextPage=True,
                )
            )

        def uses_slice_end_as_default_for_array_length():
            c = connection_from_array_slice(
                letters[:1], dict(first=1), slice_start=0,
            )
            assert c == Connection(
                edges=[
                    Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )

        def ignores_len_of_slice_if_array_slice_length_provided():
            c = connection_from_array_slice(
                letters[:2], dict(first=2),
                array_length=2, array_slice_length=1,
            )
            assert c == Connection(
                edges=[
                    Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    hasPreviousPage=False,
                    hasNextPage=True,
                )
            )

        def uses_array_slice_length_instead_of_len_function():
            class LettersWithoutLen:
                __getitem__ = letters.__getitem__

            letters_without_len = LettersWithoutLen()

            with raises(TypeError):
                len(cast(Sequence, letters_without_len))

            with raises(TypeError):
                connection_from_array_slice(letters_without_len)

            c = connection_from_array_slice(letters_without_len, array_slice_length=5)
            assert c == Connection(
                edges=[
                    Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA='),
                    Edge(node='B', cursor='YXJyYXljb25uZWN0aW9uOjE='),
                    Edge(node='C', cursor='YXJyYXljb25uZWN0aW9uOjI='),
                    Edge(node='D', cursor='YXJyYXljb25uZWN0aW9uOjM='),
                    Edge(node='E', cursor='YXJyYXljb25uZWN0aW9uOjQ='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjQ=',
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )

        def uses_default_connection_types():
            connection = connection_from_array_slice(
                letters[:1], slice_start=0, array_length=1)
            assert isinstance(connection, Connection)
            edge = connection.edges[0]
            assert isinstance(edge, Edge)
            assert len(connection.edges) == 1
            assert edge == Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA=')
            page_info = connection.pageInfo
            assert isinstance(page_info, PageInfo)
            assert page_info == PageInfo(
                startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                endCursor='YXJyYXljb25uZWN0aW9uOjA=',
                hasPreviousPage=False,
                hasNextPage=False)

        def accepts_custom_connection_type():
            class CustomConnection:
                # noinspection PyPep8Naming
                def __init__(self, edges, pageInfo):
                    self.edges = edges
                    self.page_info = pageInfo

            connection = connection_from_array_slice(
                letters[:1], slice_start=0, array_length=1,
                connection_type=CustomConnection)
            assert isinstance(connection, CustomConnection)
            edge = connection.edges[0]
            assert isinstance(edge, Edge)
            assert len(connection.edges) == 1
            assert edge == Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA=')
            page_info = connection.page_info
            assert isinstance(page_info, PageInfo)
            assert page_info == PageInfo(
                startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                endCursor='YXJyYXljb25uZWN0aW9uOjA=',
                hasPreviousPage=False,
                hasNextPage=False)

        def accepts_custom_edge_type():
            class CustomEdge:
                def __init__(self, node, cursor):
                    self.node = node
                    self.cursor = cursor

            connection = connection_from_array_slice(
                letters[:1], slice_start=0, array_length=1,
                edge_type=CustomEdge)
            assert isinstance(connection, Connection)
            assert isinstance(connection.edges, list)
            assert len(connection.edges) == 1
            edge = connection.edges[0]
            assert isinstance(edge, CustomEdge)
            assert edge.node == 'A'
            assert edge.cursor == 'YXJyYXljb25uZWN0aW9uOjA='
            page_info = connection.pageInfo
            assert isinstance(page_info, PageInfo)
            assert page_info == PageInfo(
                startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                endCursor='YXJyYXljb25uZWN0aW9uOjA=',
                hasPreviousPage=False,
                hasNextPage=False)

        def accepts_custom_page_info_type():
            class CustomPageInfo:
                # noinspection PyPep8Naming
                def __init__(
                        self, startCursor, endCursor, hasPreviousPage, hasNextPage):
                    self.startCursor = startCursor
                    self.endCursor = endCursor
                    self.hasPreviousPage = hasPreviousPage
                    self.hasNextPage = hasNextPage

            connection = connection_from_array_slice(
                letters[:1], slice_start=0, array_length=1,
                page_info_type=CustomPageInfo)
            assert isinstance(connection, Connection)
            assert isinstance(connection.edges, list)
            assert len(connection.edges) == 1
            edge = connection.edges[0]
            assert isinstance(edge, Edge)
            assert edge == Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA=')
            page_info = connection.pageInfo
            assert isinstance(page_info, CustomPageInfo)
            assert page_info.startCursor == 'YXJyYXljb25uZWN0aW9uOjA='
            assert page_info.endCursor == 'YXJyYXljb25uZWN0aW9uOjA='
            assert page_info.hasPreviousPage is False
            assert page_info.hasNextPage is False

        def provides_deprecated_connection_from_list_slice():
            with deprecated_call():
                # noinspection PyDeprecation
                c = connection_from_list_slice(
                    letters[:1], args={}, connection_type=Connection,
                    edge_type=Edge, pageinfo_type=PageInfo,
                    slice_start=0, list_length=1, list_slice_length=1)
            assert c == Connection(
                edges=[
                    Edge(node='A', cursor='YXJyYXljb25uZWN0aW9uOjA='),
                ],
                pageInfo=PageInfo(
                    startCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    endCursor='YXJyYXljb25uZWN0aW9uOjA=',
                    hasPreviousPage=False,
                    hasNextPage=False,
                )
            )
